"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const os_1 = require("os");
const code_frame_1 = require("@babel/code-frame");
const chalk_1 = require("chalk");
const lineColumn = require("line-column");
const normalizePath = require("normalize-path");
const ts = require("typescript");
function formatDiagnostics(diagnostics, context) {
    const diags = diagnostics.map(diagnostic => {
        if (diagnostic.source === 'angular') {
            return formatAngularDiagnostic(diagnostic, context);
        }
        else {
            return formatTypeScriptDiagnostic(diagnostic, context);
        }
    });
    if (diags.length > 0) {
        return diags.join(os_1.EOL) + os_1.EOL;
    }
    else {
        return '';
    }
}
exports.formatDiagnostics = formatDiagnostics;
function formatTypeScriptDiagnostic(diagnostic, context) {
    const messageText = formatDiagnosticMessage(diagnostic.messageText, '', context);
    const { file } = diagnostic;
    let message = messageText;
    if (file != null && diagnostic.start != null) {
        const lineChar = file.getLineAndCharacterOfPosition(diagnostic.start);
        const source = file.text || diagnostic.source;
        const start = {
            line: lineChar.line + 1,
            column: lineChar.character + 1
        };
        const location = { start };
        const red = chalk_1.default.red(`🚨  ${file.fileName}(${start.line},${start.column})`);
        const messages = [`${red}\n${chalk_1.default.redBright(messageText)}`];
        if (source != null) {
            if (typeof diagnostic.length === 'number') {
                const end = file.getLineAndCharacterOfPosition(diagnostic.start + diagnostic.length);
                location.end = {
                    line: end.line + 1,
                    column: end.character + 1
                };
            }
            const frame = code_frame_1.codeFrameColumns(source, location, {
                linesAbove: 1,
                linesBelow: 1,
                highlightCode: true
            });
            messages.push(frame
                .split('\n')
                .map(str => `  ${str}`)
                .join('\n'));
        }
        message = messages.join('\n');
    }
    return message + os_1.EOL;
}
function formatAngularDiagnostic(diag, context) {
    const diagnostic = diag;
    const messageText = formatDiagnosticMessage(diagnostic.messageText, '', context);
    const { file, span } = diagnostic;
    let fileName = null;
    let source = null;
    let start = null;
    let end = null;
    if (file && typeof diagnostic.start === 'number' && typeof diagnostic.length === 'number') {
        source = file.text;
        fileName = file.fileName;
        start = lineColumn(source, diagnostic.start);
        end = lineColumn(source, diagnostic.start + diagnostic.length);
    }
    else if (span) {
        source = span.start.file.content;
        fileName = span.start.file.url;
        start = span.start;
        end = span.end;
        start.line++;
        start.col++;
        end.line++;
        end.col++;
    }
    else {
        return;
    }
    const location = {
        start: {
            line: start.line,
            column: start.col
        },
        end: {
            line: end.line,
            column: end.col
        }
    };
    const red = chalk_1.default.red(`🚨  ${fileName}(${start.line},${start.col})`);
    const messages = [`${red}\n${chalk_1.default.redBright(messageText)}`];
    const frame = code_frame_1.codeFrameColumns(source, location, {
        linesAbove: 1,
        linesBelow: 1,
        highlightCode: true
    });
    messages.push(frame
        .split('\n')
        .map(str => `  ${str}`)
        .join('\n'));
    return messages.join('\n') + os_1.EOL;
}
function replaceAbsolutePaths(message, context) {
    const contextPath = normalizePath(context);
    return message.replace(new RegExp(contextPath, 'g'), '.');
}
function formatDiagnosticMessage(diagnostic, delimiter, context) {
    return replaceAbsolutePaths(ts.flattenDiagnosticMessageText(diagnostic, delimiter), context);
}
//# sourceMappingURL=format.js.map