"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const compiler_cli_1 = require("@angular/compiler-cli");
const api_1 = require("@angular/compiler-cli/src/transformers/api");
const program_1 = require("@angular/compiler-cli/src/transformers/program");
const async_loop_1 = require("../../utils/async-loop");
const reporter_1 = require("../reporter");
const remove_decorators_1 = require("../transformers/remove-decorators");
const replace_bootstrap_1 = require("../transformers/replace-bootstrap");
const resources_1 = require("../transformers/resources");
const entry_resolver_1 = require("./entry-resolver");
class BuildLoop extends async_loop_1.AsyncLoop {
    constructor(host, config, transformers) {
        super(() => this.compile());
        this.host = host;
        this.config = config;
        this.transformers = transformers;
        this.entryFile = null;
        this.resources = {};
        this.entryModule = null;
        this.program = undefined;
        this.firstRun = true;
        this.shouldEmit = false;
    }
    getProgram() {
        return this.program;
    }
    getResources(file) {
        return this.resources[file] || [];
    }
    createProgramIfNeeded() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { host } = this;
            const { changedFiles } = host.store;
            let { program } = this;
            if (changedFiles.length > 0 || this.firstRun || !program) {
                const config = compiler_cli_1.readConfiguration(this.config.project);
                const { options, rootNames } = config;
                program = program_1.createProgram({ rootNames, options, host, oldProgram: program });
                this.program = program;
                this.shouldEmit = true;
                const resource = changedFiles.find(file => !/\.ts$/.test(file));
                if (resource) {
                    Object.keys(host.store['sources']).forEach(key => {
                        if (!/node_modules/.test(key)) {
                            delete host.store['sources'][key];
                        }
                    });
                }
                changedFiles.splice(0);
                yield program.loadNgStructureAsync();
                this.updateResources(program);
            }
            return program;
        });
    }
    compile() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const program = yield this.createProgramIfNeeded();
            if (!this.shouldEmit) {
                return;
            }
            this.shouldEmit = false;
            const getTypeChecker = () => program.getTsProgram().getTypeChecker();
            const transformers = [remove_decorators_1.removeDecorators(getTypeChecker)];
            const diagnostics = [];
            diagnostics.push(...program.getNgStructuralDiagnostics());
            if (this.firstRun) {
                this.firstRun = false;
                diagnostics.push(...program.getNgOptionDiagnostics(), ...program.getTsOptionDiagnostics());
            }
            const { entryFile, host } = this;
            let { entryModule } = this;
            if (!entryModule && entryFile) {
                try {
                    const [path, className = 'default'] = entry_resolver_1.resolveEntryModuleFromMain(entryFile, host, program.getTsProgram())
                        .split('#');
                    entryModule = { className, path };
                    this.entryModule = entryModule;
                }
                catch (_) {
                    entryModule = null;
                }
            }
            if (entryModule) {
                // TODO: this has to be improved or removed
                transformers.push(replace_bootstrap_1.replaceBootstrap(file => file === entryFile, () => entryModule, getTypeChecker));
            }
            const result = program.emit({
                emitFlags: api_1.EmitFlags.All,
                customTransformers: {
                    beforeTs: [...this.transformers, ...transformers]
                }
            });
            diagnostics.push(...program
                .getTsSemanticDiagnostics()
                .filter(diag => typeof diag.messageText === 'string'
                ? !/^Module '"|'[^('|")]+\.ngfactory'|"' has no exported member '[^']+NgFactory'./.test(diag.messageText)
                : true), ...program.getTsSyntacticDiagnostics(), ...program.getNgSemanticDiagnostics(), ...result.diagnostics);
            reporter_1.reportDiagnostics(diagnostics);
        });
    }
    updateResources(program) {
        program
            .getTsProgram()
            .getSourceFiles()
            .filter(({ text }) => /(templateUrl)|(styleUrls)/.test(text))
            .forEach(sourceFile => this.resources[sourceFile.fileName] = resources_1.findResources(sourceFile)
            .map(({ path }) => typeof path === 'string' ? [path] : path)
            .reduce((a, b) => a.concat(b), []));
    }
}
exports.BuildLoop = BuildLoop;
//# sourceMappingURL=build-loop.js.map