"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const ts = require("typescript");
const ast_helpers_1 = require("./ast-helpers");
const interfaces_1 = require("./interfaces");
const make_transform_1 = require("./make-transform");
function replaceResources(shouldTransform) {
    const standardTransform = function (sourceFile) {
        const ops = [];
        if (!shouldTransform(sourceFile.fileName)) {
            return ops;
        }
        const resources = findResources(sourceFile);
        if (resources.length > 0) {
            // Add the replacement operations.
            ops.push(...(resources.map(resource => {
                if (resource.type === 'template') {
                    const propAssign = ts.createPropertyAssignment('template', createPreProcessorNode(resource.path, resource.type));
                    return new interfaces_1.ReplaceNodeOperation(sourceFile, resource.node, propAssign);
                }
                else if (resource.type === 'style') {
                    const literals = ts.createArrayLiteral(resource.path.map(path => createPreProcessorNode(path, resource.type)));
                    const propAssign = ts.createPropertyAssignment('styles', literals);
                    return new interfaces_1.ReplaceNodeOperation(sourceFile, resource.node, propAssign);
                }
                throw new Error('invariant error');
            })));
            // If we added a require call, we need to also add typings for it.
            // The typings need to be compatible with node typings, but also work by themselves.
            // interface NodeRequire {(id: string): any;}
            const nodeRequireInterface = ts.createInterfaceDeclaration([], [], 'NodeRequire', [], [], [
                ts.createCallSignature([], [
                    ts.createParameter([], [], undefined, 'id', undefined, ts.createKeywordTypeNode(ts.SyntaxKind.StringKeyword))
                ], ts.createKeywordTypeNode(ts.SyntaxKind.AnyKeyword))
            ]);
            // declare var require: NodeRequire;
            const varRequire = ts.createVariableStatement([ts.createToken(ts.SyntaxKind.DeclareKeyword)], [ts.createVariableDeclaration('require', ts.createTypeReferenceNode('NodeRequire', []))]);
            ops.push(new interfaces_1.AddNodeOperation(sourceFile, ast_helpers_1.getFirstNode(sourceFile), nodeRequireInterface));
            ops.push(new interfaces_1.AddNodeOperation(sourceFile, ast_helpers_1.getFirstNode(sourceFile), varRequire));
        }
        return ops;
    };
    return make_transform_1.makeTransform(standardTransform);
}
exports.replaceResources = replaceResources;
/*
export function findInlineResources(sourceFile: ts.SourceFile, typeChecker: ts.TypeChecker) {
    return collectDeepNodes<ts.ObjectLiteralExpression>(sourceFile, ts.SyntaxKind.ObjectLiteralExpression)
        .map(node => collectDeepNodes<ts.PropertyAssignment>(node, ts.SyntaxKind.PropertyAssignment))
        .reduce((prev, curr) => curr ? prev.concat(curr) : prev, [])
        .map(node => ({
            node, key: getContentOfKeyLiteral(node.name)!
        }))
        .filter(({key}) => {
            if(!key) {
                return false
            }

            return key === 'template' || key === 'styles'
        })
        .map(({node, key}) => {
            if(ts.isExpressionStatement(node)) {
                typeChecker.getContextualType(node.expression)!.getProperties()[0].valueDeclaration
            }
        })
}*/
function findResources(sourceFile) {
    // Find all object literals.
    return ast_helpers_1.collectDeepNodes(sourceFile, ts.SyntaxKind.ObjectLiteralExpression)
        .map(node => ast_helpers_1.collectDeepNodes(node, ts.SyntaxKind.PropertyAssignment))
        .reduce((prev, curr) => curr ? prev.concat(curr) : prev, [])
        .filter((node) => {
        const key = getContentOfKeyLiteral(node.name);
        if (!key) {
            // key is an expression, can't do anything.
            return false;
        }
        return key === 'templateUrl' || key === 'styleUrls';
    })
        .map((node) => {
        const key = getContentOfKeyLiteral(node.name);
        if (key === 'templateUrl') {
            const path = getResourceRequest(node.initializer, sourceFile);
            return { path, node, type: 'template' };
        }
        else if (key === 'styleUrls') {
            const arr = ast_helpers_1.collectDeepNodes(node, ts.SyntaxKind.ArrayLiteralExpression);
            if (!arr || arr.length === 0 || arr[0].elements.length === 0) {
                return;
            }
            return {
                type: 'style',
                path: arr[0].elements.map(element => getResourceRequest(element, sourceFile)),
                node
            };
        }
    })
        .filter(resource => resource !== undefined);
}
exports.findResources = findResources;
function getContentOfKeyLiteral(node) {
    if (!node) {
        return null;
    }
    else if (node.kind === ts.SyntaxKind.Identifier) {
        return node.text;
    }
    else if (node.kind === ts.SyntaxKind.StringLiteral) {
        return node.text;
    }
    else {
        return null;
    }
}
function getResourceRequest(element, sourceFile) {
    let path = null;
    if (element.kind === ts.SyntaxKind.StringLiteral ||
        element.kind === ts.SyntaxKind.NoSubstitutionTemplateLiteral) {
        const url = element.text;
        // If the URL does not start with ./ or ../, prepends ./ to it.
        path = `${/^\.?\.\//.test(url) ? '' : './'}${url}`;
    }
    else {
        // if not string, just use expression directly
        path = element.getFullText(sourceFile);
    }
    const directory = path_1.dirname(sourceFile.fileName);
    return path_1.resolve(directory, path);
}
function createPreProcessorNode(path, type) {
    const base64 = new Buffer(path).toString('base64');
    return ts.createLiteral(`_PRAGMA_PARCEL_TYPESCRIPT_PLUGIN_PREPROCESS_${type.toUpperCase()}(${base64})`);
}
//# sourceMappingURL=resources.js.map